package ssw.mj.test;

import java.io.StringReader;

import ssw.mj.Messages;
import ssw.mj.Parser;
import ssw.mj.Scanner;

/**
 * Test cases for the <code>Parser</code> class.
 *
 * @author AW
 */
public class ParserTest extends IdentifiedTestCase {
	private static final String CR = "\r";
	private static final String LF = "\n";

	private static TestPrintWriter output;

	//---------- helpers ----------

	private void initParser (String program) {
		output = new TestPrintWriter();
		Scanner.init(new StringReader(program), null);
	}

	//--------------------------------
	//---------- test cases ----------

	/** Test the shortest syntactically valid program: */
	public void testShortestProgram () {
		initParser("program Shortest { }");

		Parser.parse(output);
		output.verify();
	}

	public void testDoNothingProgram () {
		initParser(	"program DoNothing" + LF +
					"\tint dummy;" + LF +
					"{" + LF +
					"\tint doNothing() { dummy=0; return dummy; }" + LF +
					"" + LF +
					"\tvoid main() { " + LF +
					"\t\tdummy= doNothing();" + LF +
					"\t}" + LF +
					"}");

		Parser.parse(output);
		output.verify();
	}

	public void testDoWhileSimple () {
		initParser("program DW {" + LF +
		           "  void main () int x; {" + LF +
		           "    do x = x + 1;" + LF +
		           "    while (x < 5);" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.verify();
	}

	public void testDoWhile_1And () {
		initParser("program DW {" + LF +
		           "  void main () int x; {" + LF +
		           "    do x = x + 1;" + LF +
		           "    while (x < 5 && x > 1);" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.verify();
	}

	public void testDoWhile_1Or_2And () {
		initParser(
			"program DW {" + LF +
			"  void main () int x; {" + LF +
			"    do x = x + 1;" + LF +
			"    while (x < 5 && x > 1 || x != 3 && x != 2);" + LF +
			"  }" + LF +
			"}"
		);

		Parser.parse(output);
		output.verify();
	}

	public void testRefTypeComparison () {
		initParser(
			"program RTC" + LF +
			"  int[] iarr, jarr;" + LF +
			"{" + LF +
			"  void test () {" + LF +
			"    if (iarr == null);" + LF +
			"    if (iarr != null);" + LF +
			"    if (iarr == jarr);" + LF +
			"    if (iarr != jarr);" + LF +
			"  }" + LF +
			"  void main () { }" + LF +
			"}"
		);

		Parser.parse(output);
		output.verify();
	}

	public void testTwoReturns () {
		initParser("program TR {" + LF +
		           "  void main () {}" + LF +
		           "  int getInt (int x) {" + LF +
		           "    if (x > 5) return 0;" + LF +
		           "    if (x > 0) return 1;" + LF +
		           "    return 2;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.verify();
	}

	public void testNestedLoops () {
		initParser(
				"program NL {" + LF +
				"  void main () int i; {" + LF +
				"    i = 0;" + LF +
				"    while (i < 10) {" + LF +
				"      while (i < 5) {" + LF +
				"        i++;" + LF +
				"        if (i == 3) break;" + LF +
				"      }" + LF +
				"      i++;" + LF +
				"    }" + LF +
				"  }" + LF +
				"}"
		);

		Parser.parse(output);
		output.verify();
	}

	public void testFuncAsProc () {
		initParser("program FAP {" + LF +
							 "  int getInt () { return 1; }" + LF +
							 "  void main () {" + LF +
							 "    getInt();" + LF +
							 "  }" + LF +
							 "}");

		Parser.parse(output);
		output.verify();
	}

	public void testScriptExample () {
		initParser( "program P" + LF +
		            "" + LF +
		            "  final int size = 10;" + LF +
		            "" + LF +
		/*line  5*/ "  class Table {" + LF +
		            "    int[] pos;" + LF +
		            "    int[] neg;" + LF +
		            "  }" + LF +
		            "" + LF +
		/*line 10*/ "  Table val;" + LF +
		            "{" + LF +
		            "  void main()" + LF +
		            "    int x, i;" + LF +
		            "  { /*---------- Initialize val */" + LF +
		/*line 15*/ "    val = new Table;" + LF +
		            "    val.pos = new int[size]; val.neg = new int[size];" + LF +
		            "    i = 0;" + LF +
		            "    while (i < size) {" + LF +
		            "      val.pos[i] = 0; val.neg[i] = 0;" + LF +
		/*line 20*/ "      i++;" + LF +
		            "    }" + LF +
		            "    /*---------- Read values */" + LF +
		            "    read(x);" + LF +
		            "    while (x > 0) {" + LF +
		/*line 25*/ "      if (0 <= x && x < size) {" + LF +
		            "        val.pos[x]++;" + LF +
		            "      } else if (-size < x && x < 0) {" + LF +
		            "        val.neg[-x]++;" + LF +
		            "      }" + LF +
		/*line 30*/ "      read(x);" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "}");

		Parser.parse(output);
		output.verify();
	}

	public void testAllProds () {
		initParser(	"program AllProds" + LF +
		            "  final int nPoints = 2;" + LF +
		            "  final char exclaim = '!';" + LF +
		            "" + LF +
		/*line  5*/	"  class Point {" + LF +
		            "    int x, y;" + LF +
		            "  }" + LF +
		            "" + LF +
		            "  char[] msg;" + LF +
		/*line 10*/	"  Point[] pArr;" + LF +
		            "" + LF +
		            "{" + LF +
		            "  void printError ()" + LF +
		            "    int i;" + LF +
		/*line 15*/	"  {" + LF +
		            "    /* print error message */" + LF +
		            "    i = 0;" + LF +
		            "    while(i<len(msg)){print(msg[i]);i++;}" + LF +
		            "    print(exclaim);" + LF +
		/*line 20*/	"    do {" + LF +
		            "      i--;" + LF +
		            "      if (i < 5) break;" + LF +
		            "      print(msg[i], 2);" + LF +
		            "    } while(i > 0);" + LF +
		/*line 25*/	"  }" + LF +
		            "" + LF +
		            "  int getX (Point p) { return p.x; }" + LF +
		            "" + LF +
		            "  void setX (int x, Point p) {" + LF +
		/*line 30*/	"    if (x > -10 && x <= 9 ||" + LF +
		            "        x == 4*3/4%4+12-11*(2147483647/(-2147483647-1)%2) )" + LF +
		            "      p.x = x;" + LF +
		            "    else printError();" + LF +
		            "  }" + LF +
		/*line 35*/	"" + LF +
		            "" + LF +
		            "  void setMsg () {" + LF +
		            "    msg = new char[5];" + LF +
		            "    msg[0] = 'E';" + LF +
		/*line 40*/	"    msg[1] = 'r';" + LF +
		            "    msg[2] = 'r';" + LF +
		            "    msg[3] = 'o';" + LF +
		            "    msg[4] = 'r';" + LF +
		            "  }" + LF +
		/*line 45*/	"" + LF +
		            "  void setPoints ()" + LF +
		            "    int i;" + LF +
		            "  {" + LF +
		            "    pArr = new Point[nPoints];" + LF +
		/*line 50*/	"    i = 0;" + LF +
		            "    while (i < len(pArr)) {" + LF +
		            "      pArr[i] = new Point;" + LF +
		            "      pArr[i].x = i;" + LF +
		            "      pArr[i].y = i;" + LF +
		/*line 55*/	"      i++;" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "" + LF +
		            "  void main ()" + LF +
		/*line 60*/	"    int i;" + LF +
		            "  {" + LF +
		            "    setMsg();" + LF +
		            "    setPoints();" + LF +
		            "    read(i);" + LF +
		/*line 65*/	"    if (i < nPoints) {" + LF +
		            "      print(getX(pArr[i]));" + LF +
		            "      print(pArr[i].y);" + LF +
		            "    }" + LF +
		            "    else" + LF +
		/*line 70*/	"      printError();" + LF +
		            "    print('\\n');" + LF +
		            "  }" + LF +
		            "}");

		Parser.parse(output);
		output.verify();
	}

	// -----------------------------------------------
	// ---------- test programs with errors ----------

	public void testScanErrIn1stSymbol () {
		initParser("");

		try { Parser.parse(output); } 
		catch (Error e) { 
			assertEquals(Messages.getString("PANIC_MODE"), e.getMessage()); 
		}
		output.addExpectedLine("-- line 1, col 1: ",  "INVALID_CHAR", " ''");
		output.verify();
	}

	public void testWrongStart () {
		initParser("noclass WS { }");

		try { Parser.parse(output); } 
		catch (Error e) { 
			assertEquals(Messages.getString("PANIC_MODE"), e.getMessage()); 
		}
		output.addExpectedLine("-- line 1, col 1: program expected");
		output.verify();
	}

	public void testNoProgName () {
		initParser(	"program { }");

		try { Parser.parse(output); } 
		catch (Error e) { 
			assertEquals(Messages.getString("PANIC_MODE"), e.getMessage()); 
		}
		output.addExpectedLine("-- line 1, col 9: identifier expected");
		output.verify();
	}

	public void testWrongConstDecl () {
		initParser("program WCD" + LF +
		           "  final char ch = a;" + LF +
		           "{ void main () { } }");

		try { Parser.parse(output); } 
		catch (Error e) { 
			assertEquals(Messages.getString("PANIC_MODE"), e.getMessage()); 
		}
		output.addExpectedLine("-- line 2, col 19: ", "CONST_DECL");
		output.verify();
	}

	public void testWrongGlobalDecl () {
		initParser("program WGD" + LF +
		           "  final int MAX = 10;" + LF +
		           "  123, i, x;" + LF +
		           "  char c1, c2, c3;" + LF +
		           "{ void main () { } }");

		try { Parser.parse(output); } 
		catch (Error e) { 
			assertEquals(Messages.getString("PANIC_MODE"), e.getMessage()); 
		}
		output.addExpectedLine("-- line 3, col 3: ", "INVALID_DECL");
		output.verify();
	}

	public void testWrongMethDecl () {
		initParser("program WMD {" + LF +
		           "  return main () { }" + LF +
		           "}");

		try { Parser.parse(output); } 
		catch (Error e) { 
			assertEquals(Messages.getString("PANIC_MODE"), e.getMessage()); 
		}
		output.addExpectedLine("-- line 2, col 3: ", "METH_DECL");
		output.verify();
	}

	public void testWrongStat () {
		initParser(
				"program WS {" + LF +
				"  void main () {  " + LF +
				"    if (0 > 2) ;" + LF +
				"    123;" + LF +
				"    while (5 < 6) break;" + LF +
				"    567 no start on ident;" + LF +
				"  }" + LF +
				"}"
		);

		try { Parser.parse(output); } 
		catch (Error e) { 
			assertEquals(Messages.getString("PANIC_MODE"), e.getMessage()); 
		}
		output.addExpectedLine("-- line 4, col 5: ", "INVALID_STAT");
		output.verify();
	}

	public void testWrongFactor () {
		initParser("program WF {" + LF +
		           "  void main () int i; {  " + LF +
		           "    if (i > if) print('\\n');" + LF +
		           "  }" + LF +
		           "}");

		try { Parser.parse(output); } 
		catch (Error e) { 
			assertEquals(Messages.getString("PANIC_MODE"), e.getMessage()); 
		}
		output.addExpectedLine("-- line 3, col 13: ", "INVALID_FACT");
		output.verify();
	}

	public void testWrongRelOp () {
		initParser("program WRO {" + LF +
		           "  void main () int i; {  " + LF +
		           "    if (i x 5) print('\\n');" + LF +
		           "  }" + LF +
		           "}");

		try { Parser.parse(output); } 
		catch (Error e) { 
			assertEquals(Messages.getString("PANIC_MODE"), e.getMessage()); 
		}
		output.addExpectedLine("-- line 3, col 11: ", "REL_OP");
		output.verify();
	}

	public void testWrongDesignFollow () {
		initParser("program WDF {" + LF +
		           "  void main () int i; {  " + LF +
		           "    i**;" + LF +
		           "  }" + LF +
		           "}");

		try { Parser.parse(output); } 
		catch (Error e) { 
			assertEquals(Messages.getString("PANIC_MODE"), e.getMessage()); 
		}
		output.addExpectedLine("-- line 3, col 6: ", "DESIGN_FOLLOW");
		output.verify();
	}

	public void testAllProdsWithErrors () {
		initParser(	"program AllProdsWithErros" + LF +
		            "  final int nPoints = 2;" + LF +
		            "  final exclaim = '!';" + LF +
		            "" + LF +
		/*line  5*/ "  class {" + LF +
		            "    int x, y;" + LF +
		            "  }" + LF +
		            "" + LF +
		            "  char[ msg;" + LF +
		/*line 10*/ "  Point[] pArr;" + LF +
		            "" + LF +
		            "{" + LF +
		            "  void printError ()" + LF +
		            "    int i;" + LF +
		/*line 15*/ "  {" + LF +
		            "    /* print error message */" + LF +
		            "    i = 0;" + LF +
		            "    while(i<len(msg)){print(msg[i]);i++;}" + LF +
		            "    print(exclaim);" + LF +
		/*line 20*/ "    do {" + LF +
		            "      i--;" + LF +
		            "      if (i >= 5) break;" + LF +
		            "      print(msg[i]);" + LF +
		            "    } while(i > 0);" + LF +
		/*line 25*/ "  }" + LF +
		            "" + LF +
		            "  int getX (Point p) { return p.x; }" + LF +
		            "" + LF +
		            "  void setX (int x, Point p) {" + LF +
		/*line 30*/ "    if (x > -10 && x <= 9 ||" + LF +
		            "        x == 4*3/4%4+12-11*(2147483647/(-2147483647-1)%2) )" + LF +
		            "      p.x = x;" + LF +
		            "    else printError();" + LF +
		            "  }" + LF +
		/*line 35*/ "" + LF +
		            "" + LF +
		            "  void setMsg () {" + LF +
		            "    msg = new char[6];" + LF +
		            "    msg[0] 'E';" + LF +
		/*line 40*/ "    msg[1] = 'r';" + LF +
		            "    msg[2] = 'r';" + LF +
		            "    msg[3] = 'o';" + LF +
		            "    msg[4] = 'r';" + LF +
		            "    msg[5] = '\\n';" + LF +
		/*line 45*/ "  }" + LF +
		            "" + LF +
		            "  void setPoints ()" + LF +
		            "    int i;" + LF +
		            "  {" + LF +
		/*line 50*/ "    pArr = new Point[nPoints];" + LF +
		            "    i = 0;" + LF +
		            "    while (i < len(pArr)) {" + LF +
		            "      pArr[i] = new Point;" + LF +
		            "      pArr[i].x = i;" + LF +
		/*line 55*/ "      pArr[i].y = i" + LF +
		            "      i++;" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "" + LF +
		/*line 60*/ "  int main ()" + LF +
		            "    int i;" + LF +
		            "  {" + LF +
		            "    setMsg();" + LF +
		            "    setPoints();" + LF +
		/*line 65*/ "    read(i);" + LF +
		            "    if (i < nPoints) {" + LF +
		            "      print(getX(pArr[i]));" + LF +
		            "      print(pArr[i].y);" + LF +
		            "    }" + LF +
		/*line 70*/ "    else" + LF +
		            "      printError();" + LF +
		            "    print('\\n');" + LF +
		            "  }" + LF +
		            "}");

		try { Parser.parse(output); } 
		catch (Error e) { 
			assertEquals(Messages.getString("PANIC_MODE"), e.getMessage()); 
		}
		output.addExpectedLine("-- line 3, col 17: identifier expected");
		output.verify();
	}

	public void testNoEof () {
		initParser( "program X { void main () {} } more text");

		try { Parser.parse(output); } 
		catch (Error e) { 
			assertEquals(Messages.getString("PANIC_MODE"), e.getMessage()); 
		}
		output.addExpectedLine("-- line 1, col 31: ", "NO_EOF");
		output.verify();
	}

	// -----------------------------------------------

	public static void main (String[] args) {
		junit.textui.TestRunner.run(ParserTest.class);
	}
}
